"""Module level docstring.

"""

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

from pathlib import Path
from typing import List, Set

from digiarch.internals import IGNORED_EXTS, ArchiveFile
from tqdm import tqdm

# -----------------------------------------------------------------------------
# Function Definitions
# -----------------------------------------------------------------------------


def grouping(files: List[ArchiveFile], save_path: Path) -> None:
    """Groups files per file extension.

    Parameters
    ----------
    data_file : str
        File from which data is read.
    save_path : str
        Path to save results to
    """

    # Initialise variables
    ignored_group: List[str] = []
    exts: Set[str] = {file.ext() for file in files}

    # Create new folder in save path
    save_path = save_path / "grouped_files"
    save_path.mkdir(exist_ok=True)

    # Group files per file extension.
    for ext in tqdm(exts, desc="Grouping files"):
        group_out: Path = save_path / f"{ext}_files.txt"
        file_group: List[str] = []
        for file in files:
            if file.ext() in IGNORED_EXTS:
                ignored_group.append(str(file.path))
            if file.ext() == ext:
                file_group.append(str(file.path))
        group_out.write_text("\n".join(file_group), encoding="utf-8")

    # Save ignored files
    ignored_out: Path = save_path / "ignored_files.txt"
    ignored_out.write_text("\n".join(ignored_group), encoding="utf-8")



"""Module level docstring.

"""

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

import json
from pathlib import Path
from typing import List

from digiarch.internals import ArchiveFile
from tqdm import tqdm

# -----------------------------------------------------------------------------
# Function Definitions
# -----------------------------------------------------------------------------


def fix_extensions(files: List[ArchiveFile]) -> None:
    map_path = Path(__file__).parents[1] / "_data" / "ext_map.json"
    ext_map = json.load(map_path.open(encoding="utf-8"))
    for file in tqdm(files, desc="Fixing file extensions"):
        if file.identification:
            warning = file.identification.warning or ""
            puid = file.identification.puid
            if "Extension mismatch" in warning and puid in ext_map:
                new_name = file.path.with_name(
                    f"{file.name()}.{ext_map[puid]}"
                )
                file.path.rename(new_name)



"""Implements data classes and related utilities used throughout
Digital Archive.

"""

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

import inspect
import json
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Set

from pydantic import Field, root_validator, validator

import digiarch
from acamodels import ACABase, ArchiveFile
from natsort import natsorted

# -----------------------------------------------------------------------------
# Globals
# -----------------------------------------------------------------------------

IGNORED_EXTS: Set[str] = json.load(
    (
        Path(inspect.getfile(digiarch)).parent / "_data" / "blacklist.json"
    ).open()
).keys()

# -----------------------------------------------------------------------------
# Classes
# -----------------------------------------------------------------------------


# Metadata
# --------------------


class Metadata(ACABase):
    """Data class for keeping track of metadata used in data.json"""

    last_run: datetime
    processed_dir: Path
    file_count: Optional[int] = None
    total_size: Optional[str] = None
    duplicates: Optional[int] = None
    identification_warnings: Optional[int] = None
    empty_subdirs: Optional[List[Path]] = None
    several_files: Optional[List[Path]] = None


# File Data
# --------------------


class FileData(ACABase):
    """Data class collecting Metadata and list of FileInfo"""

    metadata: Metadata
    files: List[ArchiveFile] = []
    digiarch_dir: Path = Field(None)
    json_file: Path = Field(None)

    @root_validator
    def create_directories(cls, fields: Dict[Any, Any]) -> Dict[Any, Any]:
        metadata = fields.get("metadata")
        digiarch_dir = fields.get("digiarch_dir")
        json_file = fields.get("json_file")
        if digiarch_dir is None and metadata:
            digiarch_dir = metadata.processed_dir / "_digiarch"
            digiarch_dir.mkdir(exist_ok=True)
            fields["digiarch_dir"] = digiarch_dir
        if json_file is None and digiarch_dir:
            data_dir = digiarch_dir / ".data"
            data_dir.mkdir(exist_ok=True)
            json_file = data_dir / "data.json"
            json_file.touch()
            fields["json_file"] = json_file
        return fields

    @validator("digiarch_dir")
    def check_digiarch_dir(cls, digiarch_dir: Path) -> Path:
        if not digiarch_dir.is_dir():
            raise ValueError("Invalid digiarch directory path")
        return digiarch_dir

    @validator("json_file")
    def check_json_file(cls, json_file: Path) -> Path:
        if not json_file.is_file():
            raise ValueError("Invalid JSON file path")
        return json_file

    def dump(self) -> None:
        data = super().json(indent=2, ensure_ascii=False)
        self.json_file.write_text(data, encoding="utf-8")


# Utility
# --------------------


class DataJSONEncoder(json.JSONEncoder):
    """DataJSONEncoder subclasses JSONEncoder in order to handle
    encoding of data classes."""

    # pylint does not like this subclassing, even though it's the recommended
    # method. So we disable the warnings.

    # pylint: disable=method-hidden,arguments-differ
    def default(self, obj: object) -> Any:
        """Overrides the JSONEncoder default.

        Parameters
        ----------
        obj : object
            Object to encode.
        Returns
        -------
        dataclasses.asdict(obj) : dict
            If the object given is a data class, return it as a dict.
        super().default(obj) : Any
            If the object is not a data class, use JSONEncoder's default and
            let it handle any exception that might occur.
        """
        if isinstance(obj, Path):
            return str(obj)
        return super().default(obj)


# pylint: enable=method-hidden,arguments-differ


# -----------------------------------------------------------------------------
# Function Definitions
# -----------------------------------------------------------------------------


def size_fmt(size: float) -> str:
    """Formats a file size in binary multiples to a human readable string.

    Parameters
    ----------
    size : float
        The file size in bytes.

    Returns
    -------
    str
        Human readable string representing size in binary multiples.
    """
    for unit in ["B", "KiB", "MiB", "GiB", "TiB"]:
        if size < 1024.0:
            break
        size /= 1024.0
    return f"{size:.1f} {unit}"


def to_json(data: object, file: Path) -> None:
    """Dumps JSON files given data and a file path
    using :class:`~digiarch.data.DataJSONEncoder` as encoder.
    Output uses indent = 4 to get pretty and readable files.
    `ensure_ascii` is set to `False` so we can get our beautiful Danish
    letters in the output.

    Parameters
    ----------
    data : object
        The data to dump to the JSON file.
    dump_file: str
        Path to the file in which to dump JSON data.
    """

    with file.open("w", encoding="utf-8") as f:
        json.dump(data, f, indent=4, cls=DataJSONEncoder, ensure_ascii=False)


def natsort_path(file_list: List[ArchiveFile]) -> List[ArchiveFile]:
    """Naturally sort a list of FileInfo objects by their paths.

    Parameters
    ----------
    file_list : List[FileInfo]
        The list of FileInfo objects to be sorted.

    Returns
    -------
    List[FileInfo]
        The list of FileInfo objects naturally sorted by their path.
    """

    sorted_file_list: List[ArchiveFile] = natsorted(
        file_list, key=lambda archive_file: str(archive_file.path)
    )

    return sorted_file_list



"""This module implements checksum generation and duplicate detection.

"""

# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

import hashlib
from collections import Counter
from multiprocessing import Pool
from pathlib import Path
from typing import Any, Dict, ItemsView, List, Optional, Set

from digiarch.internals import ArchiveFile, natsort_path, to_json
from tqdm import tqdm

# -----------------------------------------------------------------------------
# Function Definitions
# -----------------------------------------------------------------------------


def file_checksum(file: Path) -> str:
    """Calculate the checksum of an input file using BLAKE2.

    Parameters
    ----------
    file : Path
        The file for which to calculate the checksum. Expects a `pathlib.Path`
        object.

    Returns
    -------
    str
        The hex checksum of the input file.

    """

    checksum: str = ""
    hasher: Any = hashlib.sha256()

    if file.is_file():
        with file.open("rb") as f:
            hasher.update(f.read())
            checksum = hasher.hexdigest()

    return checksum


def checksum_worker(file_info: ArchiveFile) -> ArchiveFile:
    """Worker used when multiprocessing checksums of FileInfo objects.

    Parameters
    ----------
    fileinfo : FileInfo
        The FileInfo object that must be updated with a new checksum value.

    Returns
    -------
    FileInfo
        The FileInfo object with an updated checksum value.
    """

    checksum: Optional[str] = file_checksum(file_info.path) or None
    updated_file_info: ArchiveFile = file_info.copy(
        update={"checksum": checksum}
    )
    return updated_file_info


def generate_checksums(files: List[ArchiveFile]) -> List[ArchiveFile]:
    """Multiprocesses a list of FileInfo object in order to assign
    new checksums.

    Parameters
    ----------
    files : List[FileInfo]
        List of FileInfo objects that need checksums.

    Returns
    -------
    List[FileInfo]
        The updated list of FileInfo objects.
    """

    # Assign variables
    updated_files: List[ArchiveFile] = []

    # Multiprocess checksum generation
    pool = Pool()
    try:
        updated_files = list(
            tqdm(
                pool.imap_unordered(checksum_worker, files),
                desc="Generating checksums",
                unit=" files",
                total=len(files),
            )
        )
    except (KeyboardInterrupt, Exception):
        pool.terminate()
        raise
    finally:
        pool.close()
        pool.join()

    return natsort_path(updated_files)


def check_collisions(checksums: List[str]) -> Set[str]:
    """Checks checksum collisions given a list of checksums as strings.
    Returns a set of collisions if any such are found.

    Parameters
    ----------
    checksums : List[str]
        List of checksums that must be checked for collisions.

    Returns
    -------
    Set[str]
        A set of colliding checksums. Empty if none are found.
    """
    checksum_counts: ItemsView[str, int] = Counter(checksums).items()
    collisions: Set[str] = set()

    for checksum, count in checksum_counts:
        if count > 1:
            # We have a collision, boys
            collisions.add(checksum)

    return collisions


def check_duplicates(files: List[ArchiveFile], save_path: Path) -> None:
    """Generates a file with checksum collisions, indicating that duplicates
    are present.

    Parameters
    ----------
    files : List[FileInfo]
        Files for which duplicates should be checked.
    save_path : Path
        Path to which the checksum collision information should be saved.
    """

    # Initialise variables
    checksums: List[str] = [
        file.checksum for file in files if file.checksum is not None
    ]
    collisions: Set[str] = check_collisions(checksums)
    file_collisions: Dict[str, str] = dict()

    for checksum in tqdm(collisions, desc="Finding duplicates"):
        hits = [
            {"name": file.name(), "path": file.path}
            for file in files
            if file.checksum == checksum
        ]
        file_collisions.update({checksum: hits})

    dups_file = save_path / "duplicate_files.json"
    to_json(file_collisions, dups_file)



"""Reporting utilities for file discovery.

"""


# -----------------------------------------------------------------------------
# Imports
# -----------------------------------------------------------------------------

from collections import Counter
from pathlib import Path
from typing import Dict, List

from acamodels import Identification
from digiarch.internals import ArchiveFile, to_json
from tqdm import tqdm

# -----------------------------------------------------------------------------
# Function Definitions
# -----------------------------------------------------------------------------


def report_results(files: List[ArchiveFile], save_path: Path) -> None:
    """Generates reports of explore_dir() results.

    Parameters
    ----------
    files: List[FileInfo]
        The files to report on.
    save_path: str
        The path in which to save the reports.

    """

    # Initialise counters & dicts
    ext_count: Counter = Counter()
    id_warnings: Dict[str, List[Dict[str, Identification]]] = dict()
    warning_key: str
    warning_count: Counter = Counter()

    # Collect information
    for file in tqdm(files, desc="Creating reports"):
        ext_count.update([file.ext()])
        if file.identification and file.identification.warning:
            if "No match" in file.identification.warning:
                warning_key = "No match"
            else:
                warning_key = file.identification.warning
            warning_list = id_warnings.get(warning_key, [])
            warning_list.append({str(file.path): file.identification.json()})
            id_warnings.update({warning_key: warning_list})

    for warning_key, warning_list in id_warnings.items():
        warning_count.update({warning_key: len(warning_list)})

    file_exts: Dict[str, int] = dict(ext_count.most_common())
    identification_warnings = {
        "counts": dict(warning_count),
        "warnings": id_warnings,
    }

    if files:
        # Create new folder in save path
        save_path = save_path / "reports"
        save_path.mkdir(exist_ok=True)

        # Save files
        to_json(file_exts, save_path / "file_extensions.json")
        to_json(
            identification_warnings,
            save_path / "identification_warnings.json",
        )
